#------------------------------------------------------
# Method for converting for converting particle concentration output from HYSPLIT in the form of the default cdump file to a  
# netCDF file formatted to meet the requirements for 3D visualization in Met3D.
#
# Summary:
# --> read HYSPLIT default particle concentration output file (cdump) into python array
# --> write to netCDF file in format required for visualization in Met3D
# 
#
# Disclaimer: This script provides a MINIMAL EXAMPLE that might serve as a useful first step by providing code snippets and the format data structures 
#
# @ Marcel Meyer (2022). UN Peace & Development Unit. Research conducted as part Visual Data Analysis Group, University of Hamburg in collaboration 
#                        with the Epidemiology Group at the University of Cambridge
#
#------------------



import datetime
import numpy as np
import pandas as pd
import os, math
import netCDF4
import socket


# Time-domain HYSPLIT output 
startdate=datetime.datetime(2010,4,14,0,0)
enddate=datetime.datetime(2010,4,15,12,0)
Delta_t=datetime.timedelta(hours=1)
iDate=startdate+Delta_t 
NumTimeSteps=36
AllTimeSteps=np.linspace(1,NumTimeSteps,NumTimeSteps)

# Spatial grid
latStart=40 
latEnd=75
lonStart=-25
lonEnd=50
Res=0.25
NumLons=int((lonEnd-lonStart)/Res+1)
NumLats=int((latEnd-latStart)/Res+1) 
AllLatitudes=np.linspace(latStart,latEnd,NumLats)
AllLongitudes=np.linspace(lonStart,lonEnd,NumLons)
AllAltitudes_mgl=[10, 250, 715, 1600, 3089, 4615, 6922, 9255, 11592, 15000] 
AllAltitudes_hPa=np.empty(len(AllAltitudes_mgl))
for i in range(len(AllAltitudes_mgl)):
    AllAltitudes_hPa[i]=1013.25*(1-(0.0065*AllAltitudes_mgl[i])/288.15)**5.255  

# initialize array to hold the HYSPLIT particle concentration data
AirConcentrationData=np.zeros([NumTimeSteps,len(AllAltitudes_hPa),len(AllLatitudes),len(AllLongitudes)])

# const for storing missing data in format required by Met.3D
MISSING_VALUE=-999E+9

# read HYSPLIT simulation data 
cdumpfile="/path/to/example/cdump.txt"
with open(cdumpfile, "r") as ConcData:
    Dummy=ConcData.readline()
    while iDate<enddate:

        # get time since start-time 
        iTimeDiff=iDate-startdate
        TimeDifference=int(iTimeDiff.days*24+iTimeDiff.seconds/(60*60))   
 
		# read data from HYSPLIT cdump file 
        iFoundData=False
        for l, line in enumerate(ConcData):
         
            # date & time 
            iYear=int(line.split()[0])
            iMonth=int(line.split()[1])
            iDay=int(line.split()[2])
            iHour=int(line.split()[3])
            iDateOfData=datetime.datetime(iYear,iMonth,iDay,iHour)

            if iDateOfData==iDate:
                
                # get lat, lon 
                iLat=float(line.split()[4]) 
                iLon=float(line.split()[5])
                
                # get indicees of position in pre-define concentration grid 
                iLatIndex=abs(math.floor((iLat-latStart)/Res))
                iLongIndex=abs(math.floor((iLon-lonStart)/Res))

                colConcStart=6 # line in cdump file  where data of interest starts 
                for i in range(len(AllAltitudes_hPa)):                    
                    iCon=float(line.split()[colConcStart+i])+float(line.split()[colConcStart+10+i])+float(line.split()[colConcStart+20+i])+float(line.split()[colConcStart+30+i])
                    AirConcentrationData[TimeDifference-1,i,iLatIndex,iLongIndex]=iCon
                iFoundData=True

            elif (iDateOfData!=iDate) and (iFoundData):
                break
                           
    
        # iterate time
        iDate=iDate+Delta_t
        print(iDate)

## write HYSPLIT output to NetCDF file in format required for processing / visualization in MET3D
                    
# define filename & open for writing 
datestring=iDate.strftime('%Y%m%d%H%M')
PathToFolderWithNCFiles="/path/to/example"
iOutFilename=os.path.join(PathToFolderWithNCFiles,'AirConcentration_HysplitExample.nc')
ncfile = netCDF4.Dataset(iOutFilename, mode="w", format="NETCDF4")

# write global NetCDF attributes and dimensions
ncfile.title="air concentration from Hysplit dispersion run"
TimeDim  = ncfile.createDimension("time", NumTimeSteps)
LatDim = ncfile.createDimension("lat", len(AllLatitudes))
LongDim = ncfile.createDimension("lon", len(AllLongitudes))
VertDim = ncfile.createDimension("plev", len(AllAltitudes_hPa))

# write  coordinate variables (time, lat, lon , height).
# time
timeVar = ncfile.createVariable("time", "d", ("time",))  
timeVar.standard_name = "time"
timeVar.long_name = "time"
timeVar.units = startdate.strftime("hours since %Y-%m-%d %H:%M:%S")
timeVar.axis = "T"
timeVar.calendar = "propleptic_gregorian"

# longitude
lonVar = ncfile.createVariable("lon", "d", ("lon"))
lonVar.standard_name = "longitude"
lonVar.long_name = "longitude"
lonVar.units = "degrees_east"
lonVar.axis = "X"

# latitude
latVar = ncfile.createVariable("lat", "f", ("lat"))
latVar.standard_name = "latitude"
latVar.long_name = "latitude"
latVar.units = "degrees_north"
latVar.axis = "Y"

# pressure levels
presVar = ncfile.createVariable("plev", "f", ("plev"))
presVar.standard_name = "air_pressure"
presVar.units = "hPa"
presVar.axis = "Z"
presVar.positive = "down"

# concentration data
concVar = ncfile.createVariable("u", "f", ("time","plev","lat","lon"))
concVar.standard_name = "air-concentration"
concVar.long_name = "air-concentration volc ash"
concVar.units = "ug/m3"

# write  coordinate data
# time, lats, longs, altitude 
timeVar[:] = AllTimeSteps
latVar[:]  = AllLatitudes
lonVar[:]  = AllLongitudes
presVar[:] = AllAltitudes_hPa.T

# write concentration data 
concVar[:,:,:,:] = AirConcentrationData

# close
ncfile.close()
