#------------------------------------------------------
# Method for for converting a single HYSPLIT locust trajectory to a  
# netCDF file with appropriate array structure for 3D trajectory visualization in Met3D
#
# Summary:
# --> read HYSPLIT DL trajectory (minimal example; as sent by Mark Cohen, NOAA) 
# --> write to netCDF file 
#
# Disclaimer: This script provides a MINIMAL EXAMPLE that might serve as a useful first step by providing code snippets and the format data structures 
#
# @ Marcel Meyer (2022). UN Peace & Development Unit. Research conducted as part Visual Data Analysis Group, University of Hamburg in collaboration 
#                        with the Epidemiology Group at the University of Cambridge
#
#------------------



# imports
import copy, time, os, glob, sys, re, math, csv
import datetime, dateutil, time 
import numpy as np
import netCDF4
import socket


# File/folder-paths 
ProjectFolder="/path/to/example"
PathToFolderTxtFiles=os.path.join(ProjectFolder,"swarm_override_08_20_40/")
PathToFolderWithNCFiles=os.path.join(ProjectFolder,"HysplitTraj_ncFiles/")
if not os.path.exists(PathToFolderWithNCFiles): os.makedirs(PathToFolderWithNCFiles)

# specs & init data structures that will hold the HYSPLIT trajectory data  
NumEnsembleMembers=1
NumParticles=1
NumDays=3
MISSING_VALUE=-999E+9
ListTimes=[]
ListTimeDifference=[]
ListLatitudes=[]
ListLongitudes=[]
ListAltitudes_m=[]
ListAltitudes_hPa_from_m=[]
ListPressureLevels_hPa=[]

# read the HYSPLIT trajectory data 
nt=1
while nt < NumDays + 1:  
    NameOutputFilename=PathToFolderTxtFiles+'swarm_override_08_20_40_tdump.'+str(nt)+".full"
    print("reading: ",NameOutputFilename)
    with open(NameOutputFilename, "r") as NAMEOutputFile:
        
        # read header lines; not required for minimal example 
        Dummy=NAMEOutputFile.readline()
        Dummy=NAMEOutputFile.readline()
        Dummy=NAMEOutputFile.readline()
        Dummy=NAMEOutputFile.readline()
        Dummy=NAMEOutputFile.readline()
        if nt == 1:
            Dummy=NAMEOutputFile.readline()
       
        for i, line in enumerate(NAMEOutputFile):
            if i == 0 and nt == 1 :
                
                # release time
                iYear = '2020'
                iMonth = line.split('    ')[4]
                iDay=line.split('    ')[5]
                iHour=line.split('    ')[6]
                iMinute=line.split('    ')[7]
                releasetime=datetime.datetime(int(iYear),int(iMonth),int(iDay),int(iHour),int(iMinute))
                ListTimes.append(releasetime)
                iTimeDiff=releasetime-releasetime
                TestTimeD=iTimeDiff.days*24*60+iTimeDiff.seconds/(60)   
                ListTimeDifference.append(TestTimeD)

                # release position
                ListLatitudes.append(float(line.split('   ')[10]))
                ListLongitudes.append(float(line.split('   ')[11]))
                iAltitude_masl=float(line.split('   ')[12])
                ListAltitudes_m.append(iAltitude_masl)
                ListAltitudes_hPa_from_m.append(1013.25*(1-(0.0065*iAltitude_masl)/288.15)**5.255) # barometric height approx.
				ListPressureLevels_hPa.append(float(line.split('   ')[13]))    

            else:          
            
                # time
                iYear =int('2020')
                iMonth =int(line.split('    ')[4])
                iDay=int(line.split('    ')[5])
                iHour=int(line.split('    ')[6])
                iMinute=int(line.split('    ')[7])       
                iTime=datetime.datetime(iYear,iMonth,iDay,iHour,iMinute)
                iTimeDiff=iTime-releasetime
                TestTimeD=iTimeDiff.days*24*60+iTimeDiff.seconds/(60)   # minutes
                ListTimeDifference.append(TestTimeD)
                ListTimes.append(iTime)                
                      
                # position
                ListLatitudes.append(float(line.split('   ')[10]))
                ListLongitudes.append(float(line.split('   ')[11]))
                iAltitude_masl=float(line.split('   ')[12])
                ListAltitudes_m.append(iAltitude_masl)
                ListAltitudes_hPa_from_m.append(1013.25*(1-(0.0065*iAltitude_masl)/288.15)**5.255) # barometric height approx.
                ListPressureLevels_hPa.append(float(line.split('   ')[13]))   

    # iterate days
    nt = nt + 1           

    
# write HYSPLIT output to NetCDF file in format required for visualization with Met.3D

# open file 
ncfilename = "%s_hysplit_locust_traj_%s.nc" % \
    (releasetime.strftime("%Y%m%d_%H%M%S"), iTime.strftime("%Y%m%d_%H%M%S"))
ncfilename = os.path.join(PathToFolderWithNCFiles, ncfilename)
ncfile = netCDF4.Dataset(ncfilename, mode="w", format="NETCDF4")

# write global NetCDF attributes (note: format of the global attributes not relevant for Met3D)
ncfile.title="Simulation data (Hysplit: locust app)"
ncfile.Conventions  = "similar to, but not exactly CF-1.6"
ncfile.CF_reference = "the following links provides some info about the format required by Met3D: https://met3d.readthedocs.io/en/latest/dataformats.html#trajectory-data-in-netcdf-format"
ncfile.featureType  = "trajectory"
now  = datetime.datetime.now()
host = socket.gethostname()
ncfile.history      = now.strftime(
    "written by convert_hysplit_to_nc_forMet3D.py at %Y-%m-%d %H:%M:%S") \
    + " on " + host

## def. dimensions 

# number of timesteps
obsDim  = ncfile.createDimension("time", len(ListTimeDifference))

# number of trajectories / particles 
trajdim = ncfile.createDimension("trajectory", NumParticles)

# number of ensemble members (Met3D can handle ensemble-data and requires specification of ensemble dim)
ensDim  = ncfile.createDimension("ensemble", NumEnsembleMembers)

## def. coordinate variables (time, lat, lon , p).

# time
timeVar = ncfile.createVariable("time", "d", ("time",))  
timeVar.standard_name = "time"
timeVar.long_name = "time"
timeVar.units = releasetime.strftime("minutes since %Y-%m-%d %H:%M")
timeVar.trajectory_starttime = releasetime.strftime("%Y-%m-%d %H:%M")
timeVar.forecast_inittime = releasetime.strftime("%Y-%m-%d %H:%M")

# longitude
lonVar = ncfile.createVariable("lon", "f", ("ensemble","trajectory","time"))
lonVar.standard_name = "longitude"
lonVar.long_name = "longitude"
lonVar.units = "degrees_east"

# latitude
latVar = ncfile.createVariable("lat", "f", ("ensemble","trajectory","time"))
latVar.standard_name = "latitude"
latVar.long_name = "latitude"
latVar.units = "degrees_north"

# altitude
presVar = ncfile.createVariable("pressure", "f", ("ensemble","trajectory","time"))
presVar.standard_name = "air_pressure"
presVar.long_name = "pressure"
presVar.units = "hPa"
presVar.positive = "down"
presVar.axis = "Z"

# time, lats, longs, altitudes
timeVar[:] = ListTimeDifference
latVar[0,0,]  = ListLatitudes
lonVar[0,0,]  = ListLongitudes
presVar[0,0,] = ListPressureLevels_hPa

# close
ncfile.close()

print( "done converting Hysplit output to nc file for Met3D\n")

