#------------------------------------------------------
# Method for converting particle concentration field (3-dimensional grid with particle material concentration per grid-cell)
# obtained from NAME simulations to netCDF files with data-format that can be processed by Met3D
#
# Summary:
# --> read NAME output 
# --| write to NetCDF file 
#
# Disclaimer: This script provides a MINIMAL EXAMPLE that might serve as a useful first step by providing code snippets and the format data structures 
#
# @ Marcel Meyer (2022). UN Peace & Development Unit. Research conducted as part Visual Data Analysis Group, University of Hamburg in collaboration 
#                        with the Epidemiology Group at the University of Cambridge
#
#------------------


# Import libs
import datetime
import numpy as np
import pandas as pd
import os, math
import netCDF4
import socket
import datetime, dateutil, time 
import numpy as np
import iris 
from scipy.ndimage import gaussian_filter

# Filepath
NAMEOutputFolder="/xyz/"

# Define 3-D regular grid to hold NAME particle concentration field output
# note: horizontal grid specs copied from NAME output field & hardcoded here for testing minimal example 
xGridOrigin=-18.63281
yGridOrigin=-12.04688
xGridSize=706
yGridSize=482
xGridResolution=0.1406250
yGridResolution=0.09375
allLatitudesConcGrid=np.zeros(yGridSize)
allLongitudesConcGrid=np.zeros(xGridSize)
for iLat in range(0,yGridSize):
    allLatitudesConcGrid[iLat]=yGridOrigin+iLat*yGridResolution 
for iLon in range(0,xGridSize):
    allLongitudesConcGrid[iLon]=xGridOrigin+iLon*xGridResolution 

# get vertical levels from NAME output file 
# note: these are similar for all NAME output files, so just take one example file
ExampleConcGridFile=os.path.join(NAMEOutputFolder,"concentration_C1_T1_202103080000.txt")
with open(ExampleConcGridFile, "r") as ConcOutputFile:
    for i, line in enumerate(ConcOutputFile):
        if i==34:
            verticalLevelLine = line.split(',')[4:54] #
            allVerticalLevelsConcGrid=np.zeros(len(verticalLevelLine)-1)
            for i in range(0,(len(verticalLevelLine)-1)):
                verticalLevelHelper=verticalLevelLine[i].split('=')
                verticalLevelHelper=float(verticalLevelHelper[1].strip('Pa'))
                allVerticalLevelsConcGrid[i]=verticalLevelHelper
allVerticalLevelsConcGrid_hPa=allVerticalLevelsConcGrid/100

# define time-domain (covering NAME run; copied from NAME input/output)
T0_Year=2021
T0_Month=3
T0_Day=8
T0_Hour=0
T0_Minute=0
T0_Second=0
DeltaT_Minutes=30 
startdate=datetime.datetime(T0_Year, T0_Month, T0_Day, T0_Hour, T0_Minute, T0_Second) 
DeltaT_Minutes=30 
Delta_t=datetime.timedelta(minutes=DeltaT_Minutes)
NumTimeSteps=49  
NumDays=1 
NumTimeStepsPerDay=49

# ensemble members. Note: Met.3D can handle ensemble data, but we just use one set of simulation results here
NumEnsembleMembers=1   

# read NAME output data 
nTimeStep=0
allDays=np.linspace(1,NumDays,NumDays) 
for d in allDays:
    
    # initialize array for holding times of this day
    Times=np.zeros((NumTimeStepsPerDay,))  
	
    nt=0
    while nt < NumTimeStepsPerDay:
    
        # initialize arrays for holding a single timestep of the NAME concentration field
        iConcField=np.zeros([len(allVerticalLevelsConcGrid),len(allLatitudesConcGrid),len(allLongitudesConcGrid)])
        iConcFieldLogScaled=np.zeros([len(allVerticalLevelsConcGrid),len(allLatitudesConcGrid),len(allLongitudesConcGrid)])
        iConcFieldSmoothed1Sigma=np.zeros([len(allVerticalLevelsConcGrid),len(allLatitudesConcGrid),len(allLongitudesConcGrid)])
        iConcFieldSmoothedLogScaled=np.zeros([len(allVerticalLevelsConcGrid),len(allLatitudesConcGrid),len(allLongitudesConcGrid)])
          
        # set date-time of this time-step 
        if d==1 and nt==0:
            iTimeObj=startdate 
        else:
            iTimeObj=startdate + nTimeStep*Delta_t
        datestring=iTimeObj.strftime('%Y%m%d%H%M')
        Time_infilename=datestring + '00'
        print("processing: ", datestring)
        nt=nt+1
        nTimeStep=nTimeStep+1
        print(nt,nTimeStep)
        
        # set index for defining NAME filename
        index=str(nTimeStep)

        # get relative time for time-coord array
        iTimeDiff=iTimeObj-startdate
        iTimeDiffInHours=iTimeDiff.days*24+iTimeDiff.seconds/(60*60)   
        Times[nt-1]=iTimeDiffInHours

        # def. NAME file-name
        iConcGridFile=os.path.join(NAMEOutputFolder,'concentration_C1_T'+index+'_'+datestring+'.txt')
        print("processing:", iConcGridFile)

        with open(iConcGridFile, "r") as iConcFile:
            for i, line in enumerate(iConcFile):
                if i>36:
                    
                    # get (lat,lon) indicees of conc. entry
                    # note: NAME indexing starts at 1, Python inexing at 0
                    xIndex = int(line.split(',')[0])-1
                    yIndex = int(line.split(',')[1])-1
                    iLonConcGrid=float(line.split(',')[2])
                    iLatConcGrid=float(line.split(',')[3])

                    # take vertical level from output file
                    iVerticalLevels=allVerticalLevelsConcGrid    
                   
                    # iterate through all verticel levels
                    for iLev in range(0,len(iVerticalLevels)):

                        # get altitude of this level 
                        iAltitude_hPa=iVerticalLevels[iLev]/100
                        
                        # find index of altitude in array with all vertical levels
                        for verticalIndex in range(0,len(allVerticalLevelsConcGrid)-1):                            
                            
                            # vertical levels in units of pressure decreases with increasing altitude 
                            iLevel=allVerticalLevelsConcGrid[verticalIndex+1]/100
                            iLevelAbove=allVerticalLevelsConcGrid[verticalIndex]/100

                            if ((iAltitude_hPa>=iLevelAbove) and (iAltitude_hPa<iLevel)):
                                
                                # read particle materical concentration  
                                iConcValue=float(line.split(',')[4+iLev])

                                # note: 
								# the NAME default unit for the particle material concentration is [g/m3],
                                # where "g" does not correspond to a physical mass, but is a placeholder for a user-defined material mass 
								# (see NAME user-guide). As the grid-resolution is usually defined on scales of kilometers for our use-cases,
								# we scale from g / m3 to g / km3 here. As the particle material at release is set to unit release (1 g / m3), 
								# we are not dealing with actual physcial quantities (as e.g. numbers of locusts), and the re-scaling is for convencience only
                                scaleFactorConcValues=1000000000
                    
                                # write concentration value and log-scale for visualization
                                if iConcValue!=0:
                                    iConcValueLogScaled=math.log10(iConcValue*scaleFactorConcValues)
                                    iConcValueScaled=iConcValue*scaleFactorConcValues                                    
                                    iConcFieldLogScaled[verticalIndex][yIndex][xIndex]=iConcValueLogScaled
                                    iConcField[verticalIndex][yIndex][xIndex]=iConcValueScaled

     
        # Gaussian smoothing of the field for testing visualization
        iConcFieldSmoothed1Sigma=gaussian_filter(iConcField, sigma=1)  
        iConcFieldSmoothedLogScaled=gaussian_filter(iConcFieldLogScaled, sigma=1)  
                        
                        
        ### write to NetCDF file
		
		# open file for writing
        ConcOutputFolder="/xyz/"
        ncfilename = "SwarmPDFApproximation_%s.nc" % (datestring)
        ncfilename = os.path.join(ConcOutputFolder, ncfilename) 
        ncfile = netCDF4.Dataset(ncfilename, mode="w", format="NETCDF4")

        # write global NetCDF attributes & dimensions
        ncfile.title="3-D field of swarm air-concentration"
        TimeDim  = ncfile.createDimension("time", 1)
        LatDim = ncfile.createDimension("lat", len(allLatitudesConcGrid))
        LongDim = ncfile.createDimension("lon", len(allLongitudesConcGrid))
        VertDim = ncfile.createDimension("plev", len(allVerticalLevelsConcGrid))

        # write coordinate variables (time, lat, lon , height).
        
        # time
        timeVar = ncfile.createVariable("time", "d", ("time",))  
        timeVar.standard_name = "time"
        timeVar.long_name = "time"
        timeVar.units = startdate.strftime("hours since %Y-%m-%d %H:%M:%S")
        timeVar.axis = "T"
        timeVar.calendar = "propleptic_gregorian"

        # longitude
        lonVar = ncfile.createVariable("lon", "d", ("lon"))
        lonVar.standard_name = "longitude"
        lonVar.long_name = "longitude"
        lonVar.units = "degrees_east"
        lonVar.axis = "X"

        # latitude
        latVar = ncfile.createVariable("lat", "f", ("lat"))
        latVar.standard_name = "latitude"
        latVar.long_name = "latitude"
        latVar.units = "degrees_north"
        latVar.axis = "Y"

        # pressure levels
        presVar = ncfile.createVariable("plev", "f", ("plev"))
        presVar.standard_name = "air_pressure"
        presVar.units = "hPa"
        presVar.axis = "Z"
        presVar.positive = "down"

        # particle material concentration
        concVar = ncfile.createVariable("SwarmAirConc", "f", ("time","plev","lat","lon"))
        concVar.standard_name = "SwarmAirConc"
        concVar.long_name = "SwarmAirConc"
        concVar.units = "g/km3"

        # smoothed particle material concentration
        concVarSmooth1S = ncfile.createVariable("SwarmAirConcSmoothed", "f", ("time","plev","lat","lon"))
        concVarSmooth1S.standard_name = "SwarmAirConc smoothed 1S"
        concVarSmooth1S.long_name = "SwarmAirConc smoothed 1 Sigma"
        concVarSmooth1S.units = "g/km3"

		# log-scaled particle material concentration 
        concVarLogScaled = ncfile.createVariable("SwarmAirConcLogScaled", "f", ("time","plev","lat","lon"))
        concVarLogScaled.standard_name = "SwarmAirConc log scaled"
        concVarLogScaled.long_name = "SwarmAirConc log scaled"
        concVarLogScaled.units = "g/km3"

		# smoothed log-scaled particle material concentration
        concVarLogScaledSmoothed = ncfile.createVariable("SwarmAirConcLogScaleSmoothed", "f", ("time","plev","lat","lon"))
        concVarLogScaledSmoothed.standard_name = "SwarmAirConc scaled smoothed"
        concVarLogScaledSmoothed.long_name = "SwarmAirConc log scale smoothed"
        concVarLogScaledSmoothed.units = "g/km3"

        # write coordinate data
        timeVar[:] = iTimeDiffInHours
        latVar[:]  = allLatitudesConcGrid
        lonVar[:]  = allLongitudesConcGrid
        presVar[:] = allVerticalLevelsConcGrid_hPa

        # write spore concentration values 
        concVar[0,:,:,:] = iConcField
        concVarSmooth1S[0,:,:,:] = iConcFieldSmoothed1Sigma
        concVarLogScaled[0,:,:,:] = iConcFieldLogScaled
        concVarLogScaledSmoothed[0,:,:,:] = iConcFieldSmoothedLogScaled

        # close
        ncfile.close()
