#------------------------------------------------------
# Method for converting NAME particle output to .csv files for visualization in ParaView
#
# Summary:
# --> read NAME output
# --> convert vertical coordinate if necessary
# --> write to NetCDF file 
#
# Disclaimer: This script provides a MINIMAL EXAMPLE that might serve as a useful first step by providing code snippets and the format data structures 
#
# @ Marcel Meyer (2022). UN Peace & Development Unit. Research conducted as part Visual Data Analysis Group, University of Hamburg in collaboration 
#                        with the Epidemiology Group at the University of Cambridge
#
#------------------



## Imports
import copy, time, os, glob, sys, re, math, csv
import datetime, dateutil, time 
import numpy as np


## Function Defs

# Barometric height approximation
def convertFrom_masl_to_hPa(iAltitude_m):
    iAltitude_approx_hPa=1013.25*(1-(0.0065*iAltitude_m)/288.15)**5.255 
    return iAltitude_approx_hPa 

def convertFrom_hPa_to_masl(iAltitude_hPa):
    iAltitude_approx_masl=-(288.15/0.0065)*((iAltitude_hPa/1013.25)**(1/5.255)-1)
    return iAltitude_approx_masl


# File/Folder-Paths
ProjectFolder="/xyz/"
NAMEOutputFolder=ProjectFolder+"NAME Output Particle Data"
PathToFolderWithCSVFiles=os.path.join(ProjectFolder,"NAME Post-Processed Particle Data/")
if not os.path.exists(PathToFolderWithCSVFiles): os.makedirs(PathToFolderWithCSVFiles)

# Config & Constants for data from sample test-run
MISSING_VALUE=-999E+9
NumParticles=1000  
NumDays=1
NumTimeSteps=200 

## Load NAME particle output data, convert the vertical coordinate and write to particle data to .csv format required for processing with ParaView

# process all days with NAME particle output
nt=1
while nt < NumDays + 1:

    # init. arrays for storing NAME particle data, aux. data and time 
    Times=np.ones((NumTimeSteps,))*MISSING_VALUE           
    Longs=np.ones((NumParticles,NumTimeSteps))*MISSING_VALUE     
    Lats=np.ones((NumParticles,NumTimeSteps))*MISSING_VALUE       
    Ps=np.ones((NumParticles,NumTimeSteps))*MISSING_VALUE 
    Temperatures=np.ones((NumParticles,NumTimeSteps))*MISSING_VALUE
	allTimeStr=[]
   
    # process all NAME particles
    for i in range(1,1001):
    
        # NAME output file 
        NameOutputFilename=NAMEOutputFolder+"/Data_Traj_C1_Particle"+str(i)+".txt"
        print("reading: ",NameOutputFilename)    
        with open(NameOutputFilename, "r") as NAMEOutputFile:
        
            # deal with header info
            Dummy=NAMEOutputFile.readline()
            Dummy=NAMEOutputFile.readline()
            Dummy=NAMEOutputFile.readline()
            Dummy=NAMEOutputFile.readline()
            Dummy=NAMEOutputFile.readline()
            Dummy=NAMEOutputFile.readline()
            Dummy=NAMEOutputFile.readline()
            Dummy=NAMEOutputFile.readline()
            Dummy=NAMEOutputFile.readline()
            Dummy=NAMEOutputFile.readline()
            Dummy=NAMEOutputFile.readline()

            # read particle-information data line by line
            for l, line in enumerate(NAMEOutputFile):
            
                # set first time-step as release time & release position
                if l == 0:
                
                    # get release time 
					iTimeStr = line.split(",")[4].strip('UTC')
                    iDay=int(iTimeStr.split('/')[0])
                    iMonth=int(iTimeStr.split('/')[1])
                    iYear=int((iTimeStr.split('/')[2]).split(' ')[0])
                    iHour=int((iTimeStr.split('/')[2]).split(' ')[1].split(':')[0])
                    iMinute=int((iTimeStr.split('/')[2]).split(' ')[1].split(':')[1])
                    iSecond=int((iTimeStr.split('/')[2]).split(' ')[1].split(':')[2])                    
                    iReleaseTime=datetime.datetime(iYear,iMonth,iDay,iHour,iMinute,iSecond)                    
                    iTimeDiff=iReleaseTime-iReleaseTime 
                    iTimeD=iTimeDiff.days*24*60+iTimeDiff.seconds/(60)   
                    
                    # get release position                                               
                    iLat=(float(line.split(',')[7]))
                    iLon=(float(line.split(',')[6]))
                    iPres=(float(line.split(',')[22]))

                    # store release time & position 
                    Times[l]=iTimeD   
                    allTimeStr.append(iTimeStr)       
                    Longs[i-1,l]=iLon  
                    Lats[i-1,l]=iLat       
                    Ps[i-1,l]=iPres/100

                else:

                    # get particle time 
                    iTimeStr = line.split(",")[4].strip('UTC')
                    iDay=int(iTimeStr.split('/')[0])
                    iMonth=int(iTimeStr.split('/')[1])
                    iYear=int((iTimeStr.split('/')[2]).split(' ')[0])
                    iHour=int((iTimeStr.split('/')[2]).split(' ')[1].split(':')[0])
                    iMinute=int((iTimeStr.split('/')[2]).split(' ')[1].split(':')[1])
                    iSecond=int((iTimeStr.split('/')[2]).split(' ')[1].split(':')[2])
                    iTime=datetime.datetime(iYear,iMonth,iDay,iHour,iMinute,iSecond)
                    iTimeDiff=iTime-iReleaseTime
                    iTimeD=iTimeDiff.days*24*60+iTimeDiff.seconds/(60)   # minutes                    

                    # get particle position 
                    iLat=(float(line.split(',')[7]))
                    iLon=(float(line.split(',')[6]))
                    iPres=(float(line.split(',')[22]))

                    # store time & location 
                    Times[l]=iTimeD  
                    allTimeStr.append(iTimeStr)          
                    Longs[i-1,l]=iLon  
                    Lats[i-1,l]=iLat       
                    Ps[i-1,l]=iPres/100      # hPa           
                    
                    # get & store temperature along particle trajectories
                    iTemperature=float(line.split(',')[21])
                    Temperatures[i-1,l]=iTemperature

     
    # write particle info to csv file 
    # Format: one file per time-step 
    # Header-line: "X","Y","Z","auxData","time"
    for iT in range(0,NumTimeSteps):
        iTime=Times[iT]
        iTimeStr=allTimeStr[iT]
        if iTime>MISSING_VALUE:

            # open csv file for writing
            csvFilename="particleData_TimeStep_"+str(iT)+".csv"
            csvFullFilename=os.path.join(PathToFolderWithCSVFiles,csvFilename)
            csvFileObj=open(csvFullFilename,'w')
            csvFileWriter=csv.writer(csvFileObj)
            print(["writing file:", csvFilename])                   

            # write header 
            header=['X [deg.]','Y [deg.]','Z[m asl]','Z [hPa]','auxTemp [K]','time (dd/mm/yyyy :hr:mm:s [UTC])','travel duration [minutes since release]']
            csvFileWriter.writerow(header)

            # write data for all particles
            for iP in range(0,NumParticles):
                
                # get data array
                iLat=Lats[iP][iT]
                iLon=Longs[iP][iT]
                iPres=Ps[iP][iT]
                iPres_masl=convertFrom_hPa_to_masl(iPres)
                auxTemp=Temperatures[iP][iT]

                # write to csv
                iDataRow=[round(iLon,3),round(iLat,3),round(iPres_masl),round(iPres,2),round(auxTemp,2),iTimeStr,iTime]
                csvFileWriter.writerow(iDataRow)         

            csvFileObj.close()
    
    # iterate days with NAME particle outut 
    nt = nt+1    


                    
                   

                   
 