#------------------------------------------------------
# Method for converting 2-D soil moisture fields from NAME NetCDF files for processing in Met3D
#
# Summary:
#  -- read the 2-D soil moisture data from txt file 
#  -- read the surface pressure field from the UM 
#  -- put on the same grid
#  -- sea-land masking 
#  -- write to NetCDF file 
#
# Disclaimer: This script provides a MINIMAL EXAMPLE that might serve as a useful first step by providing code snippets and the format data structures 
#
# @ Marcel Meyer (2022). UN Peace & Development Unit. Research conducted as part Visual Data Analysis Group, University of Hamburg in collaboration 
#                        with the Epidemiology Group at the University of Cambridge
#
#------------------


# import libs
import datetime
import numpy as np
import pandas as pd
import os, math
import netCDF4
import socket
import datetime, dateutil, time 
import numpy as np
import glob
import iris


# file/folder paths to NAME soil moisture data 
ProjectFolder="/xyz/NAME_Locusts_May2020_Example/"
NAMEOutputFolder=ProjectFolder+"locust_fcast_20200708/"
PathToFolderWithNCFiles=os.path.join(ProjectFolder,"NetCDF_forVisualization/")

# construct 2-D grid to store NAME output
xGridOrigin=8.789062
yGridOrigin=-15.23438
xGridSize=454
yGridSize=508
xGridResolution=0.1406250
yGridResolution=9.3750000E-02
allLatitudesSMGrid=np.zeros(yGridSize)
allLongitudesSMGrid=np.zeros(xGridSize)
for iLat in range(0,yGridSize):
    allLatitudesSMGrid[iLat]=yGridOrigin+iLat*yGridResolution 
for iLon in range(0,xGridSize):
    allLongitudesSMGrid[iLon]=xGridOrigin+iLon*xGridResolution 

# define missing value and Ensemble Members - as required for processing in Met3D
MISSING_VALUE=-999E+9
NumEnsembleMembers=1

# Def. time-domain & resolution
T0_Year=2020
T0_Month=7
T0_Day=8
T0_Hour=0
T0_Minute=0
T0_Second=0
startdate=datetime.datetime(T0_Year, T0_Month, T0_Day, T0_Hour, T0_Minute, T0_Second) 
DeltaT_Days=1  
Delta_t=datetime.timedelta(days=DeltaT_Days)
NumDays=5
NumTimeStepsPerDay=1 

# Load auxiliary SST data field as helper for distinguishing land/see boundary
SST_File=netCDF4.Dataset("/xyz/ExampleSeaSurfaceTemperatureData.nc")
SST_Data=np.array(SST_File.variables['analysed_sst'][0][:][:])
latMinSST=-90
lonMinSST=-180
resSSTGrid=0.05

# load a single time-step of UM surface pressure (as Met3D plots the Earth's surface along surface presssure data)
MetDataPT2=os.path.join(ProjectFolder,"ExampleMetData/MO201803030000.UMG_Mk10_I_L59PT2.pp")
MetDataPT3=os.path.join(ProjectFolder,"ExampleMetData/MO201803030000.UMG_Mk10_I_L59PT3.pp")
MetDataPT6=os.path.join(ProjectFolder,"ExampleMetData/MO201803030000.UMG_Mk10_I_L59PT6.pp")
MetDataPT7=os.path.join(ProjectFolder,"ExampleMetData/MO201803030000.UMG_Mk10_I_L59PT7.pp")

metVarNamesI=['surface_air_pressure']

iCubeListPT2APS=iris.load(MetDataPT2,metVarNamesI)
iCubeListPT3APS=iris.load(MetDataPT3,metVarNamesI)
iCubeListPT6APS=iris.load(MetDataPT6,metVarNamesI)
iCubeListPT7APS=iris.load(MetDataPT7,metVarNamesI)

LatitudesPT2=iCubeListPT2APS[0].dim_coords[0]    
LatitudesPT6=iCubeListPT6APS[0].dim_coords[0]    
AllLatitudes=np.append(LatitudesPT6.points,LatitudesPT2.points[2:len(LatitudesPT2.points)])
LatMinAPS=min(AllLatitudes)   
resAPSGridLats=AllLatitudes[1]-AllLatitudes[0]

LongitudesPT2=iCubeListPT2APS[0].dim_coords[1]    
LongitudesPT3=iCubeListPT3APS[0].dim_coords[1] 
# construct longitude array covering all longitudes of merged PT domains
# PT 2, 6 cover longitudes -45 to 45 and PT 3,7 cover longitudes 45 - 135
# PT 2, 6 are defined using longitudes running from 315 until 414 
# PT 3, 7 are defined using longitudes running as expected from 45 to 135
# there is an overlap of two longitudes at boundary 2 and 3 as well as 6 and 7
# for merging: shift longitudes to common frame, then append counting duplicates only once
LongitudesPT2Shifted=np.empty(len(LongitudesPT2.points))
i=0
for iLon in LongitudesPT2.points:
    if iLon<=180:
        LongitudesPT2Shifted[i]=iLon
    if iLon>180:
        LongitudesPT2Shifted[i]=iLon-360
    i=i+1
AllLongitudes=np.append(LongitudesPT2Shifted,LongitudesPT3.points[2:len(LongitudesPT3.points)])
LonMinAPS=min(AllLongitudes) 
resAPSGridLons=AllLongitudes[1]-AllLongitudes[0]

# read and merge surface pressure data
SurfacePressureData=np.empty([1,len(AllLatitudes),len(AllLongitudes)])    

# fill surface pressure data
SurfacePressureData[0,0:len(LatitudesPT6.points),0:len(LongitudesPT2Shifted)]=iCubeListPT6APS[0].data
SurfacePressureData[0,0:len(LatitudesPT6.points),len(LongitudesPT2Shifted):len(LongitudesPT2Shifted)+len(LongitudesPT3.points)-2]=iCubeListPT7APS[0].data[:,2:len(LongitudesPT3.points)]
SurfacePressureData[0,len(LatitudesPT6.points):len(LatitudesPT6.points)+len(LatitudesPT2.points)-2,0:len(LongitudesPT2Shifted)]=iCubeListPT2APS[0].data[2:len(LatitudesPT2.points),:]
SurfacePressureData[0,len(LatitudesPT6.points):len(LatitudesPT6.points)+len(LatitudesPT2.points)-2,len(LongitudesPT2Shifted):len(LongitudesPT2Shifted)+len(LongitudesPT3.points)-2]=iCubeListPT3APS[0].data[2:len(LatitudesPT2.points),2:len(LongitudesPT3.points)]
    
# cut surface pressure grid to same grid as soil moisture
# define new surface pressure grid with same extent as the soil moisture grid
APSField=np.zeros([NumTimeStepsPerDay,len(allLatitudesSMGrid),len(allLongitudesSMGrid)])
  
# write surface pressure into grid with same extent as soil moisture 
for latIndex,iLatInSMGrid in enumerate(allLatitudesSMGrid):
    for lonIndex,iLonInSMGrid in enumerate(allLongitudesSMGrid):
                
        # get index in surface pressure grid
        distToLatMinAPS=iLatInSMGrid-LatMinAPS
        distToLonMinAPS=iLonInSMGrid-LonMinAPS
        iLatIndexAPS=math.floor(distToLatMinAPS/resAPSGridLats)
        iLonIndexAPS=math.floor(distToLonMinAPS/resAPSGridLons)

        # get value of surface pressure in original grid
        iSurfacePressure=SurfacePressureData[0][iLatIndexAPS][iLonIndexAPS]

        # write surface pressure value into new grid
        APSField[0][latIndex][lonIndex]=iSurfacePressure
                  

# read time-dependent soil moisture data 
allDays=np.linspace(1,NumDays,NumDays) 
for d in allDays:

    # initialize empty soil moisture grid
    SMField=np.zeros([NumTimeStepsPerDay,len(allLatitudesSMGrid),len(allLongitudesSMGrid)])
  
    # initialize array for holding times 
    Times=np.zeros((NumTimeStepsPerDay,))                
 
    nt=0
    while nt < NumTimeStepsPerDay:
    
        # set date-time of this time-step 
        if d==1 and nt==0:
            iTimeObj=startdate 
        else:
            iTimeObj=startdate + (d-1)*Delta_t
        datestring=iTimeObj.strftime('%Y%m%d%H%M')
        
        # find the file with info about soil moisture (the time here is irregular, hence glob it)
        NAMEOutputFolderThisDay=os.path.join(NAMEOutputFolder,'day'+str(int(d)),"Met_Data_Fields_C1_*")
        SMFileName=glob.glob(NAMEOutputFolderThisDay)[0]
        print("processing:", SMFileName)

        with open(SMFileName, "r") as iSMFile:
            for i, line in enumerate(iSMFile):
                if i>36:
                    
                    # get (lat,lon) indicees of dep. entry
                    # note, these start at 1 in NAME output, but python indexing starts at zero
                    xIndex = int(line.split(',')[0])-1
                    yIndex = int(line.split(',')[1])-1
                    iLonSMGrid=float(line.split(',')[2])
                    iLatSMGrid=float(line.split(',')[3])

                    # get soil moisture 
                    iSoilMoisture=float(line.split(',')[4])

                    # write into soil moisture grid
                    SMField[nt-1][yIndex][xIndex]=iSoilMoisture

        # mask to land-masses by using SST as surrogate for land-sea mask
        for latIndex,iLatInSMGrid in enumerate(allLatitudesSMGrid):
            for lonIndex,iLonInSMGrid in enumerate(allLongitudesSMGrid):
                
                # get index in SST grid
                distToLatMinSST=iLatInSMGrid-latMinSST
                distToLonMinSST=iLonInSMGrid-lonMinSST
                iLatIndexSST=math.floor(distToLatMinSST/resSSTGrid)
                iLonIndexSST=math.floor(distToLonMinSST/resSSTGrid)

                # get value of SST 
                iSST=SST_Data[iLatIndexSST][iLonIndexSST]

                # if over sea, set to missing value
                if iSST>0:
                    SMField[nt-1][latIndex][lonIndex]=MISSING_VALUE                    
              
        # iterate timesteps per day
        nt=nt+1


    ## write soil moisture into NetCDF file (1 file per day)    
    ncfilename = "SoilMoisture_APS_day2Mean_%s.nc" % (datestring)
    ncfilename = os.path.join(PathToFolderWithNCFiles, ncfilename) 
    ncfile = netCDF4.Dataset(ncfilename, mode="w", format="NETCDF4")
    ncfile.title="SoilMoistureAndAPS"

    # Define nc dimensions 
    TimeDim  = ncfile.createDimension("time", len(Times))
    LatDim = ncfile.createDimension("lat", len(allLatitudesSMGrid))
    LongDim = ncfile.createDimension("lon", len(allLongitudesSMGrid))
    
    ### Define coordinate variables (time, lat, lon, height).
    
    # time
    timeVar = ncfile.createVariable("time", "d", ("time",))  
    timeVar.standard_name = "time"
    timeVar.long_name = "time"
    timeVar.units = iTimeObj.strftime("hours since %Y-%m-%d %H:%M:%S")
    timeVar.axis = "T"
    timeVar.calendar = "propleptic_gregorian"

    # longitude
    lonVar = ncfile.createVariable("lon", "d", ("lon"))
    lonVar.standard_name = "longitude"
    lonVar.long_name = "longitude"
    lonVar.units = "degrees_east"
    lonVar.axis = "X"

    # latitude
    latVar = ncfile.createVariable("lat", "f", ("lat"))
    latVar.standard_name = "latitude"
    latVar.long_name = "latitude"
    latVar.units = "degrees_north"
    latVar.axis = "Y"

    # soil moisture
    SMVar = ncfile.createVariable("SoilMoisture", "f", ("time","lat","lon"))
    SMVar.standard_name = "SoilMoisture"
    SMVar.long_name = "SoilMoisture"
    SMVar.units = "kg/m2"
  
    # surface pressure
    APSVar = ncfile.createVariable("surfacePressure", "f", ("time","lat","lon"))
    APSVar.standard_name = "surfacePressure"
    APSVar.long_name = "surfacePressure"
    APSVar.units = "Pa"

    # write coordinate data
    timeVar[:] = Times
    latVar[:]  = allLatitudesSMGrid
    lonVar[:]  = allLongitudesSMGrid
    
    # write soil moisture and surface pressure
    SMVar[:,:,:] = SMField
    APSVar[:,:,:] = APSField

    # close
    ncfile.close()
 