#------------------------------------------------------
# Method for converting NAME topography files to NetCDF files holding topography and approximated surface pressure data that can be processed by Met3D for visualizing the terrain 
#
# Summary:
#  --> read terrain data using Python Iris (NAME topography file; UMs pp format)
#  --> convert metric altitude to approximate surface pressure, because Met3D does not have the functionality to visualize terrain data in unit of meters, but instead requires surface pressure values
#  --> write approximate surface pressure values for terrain visualization in Mt.3D
#
# Disclaimer: This script provides a MINIMAL EXAMPLE that might serve as a useful first step by providing code snippets and the format data structures 
#
# @ Marcel Meyer (2022). UN Peace & Development Unit. Research conducted as part Visual Data Analysis Group, University of Hamburg in collaboration 
#                        with the Epidemiology Group at the University of Cambridge
#
#------------------


# Import libs
import datetime
import numpy as np
import pandas as pd
import os, math
import netCDF4
import socket
import datetime, dateutil, time 
import numpy as np
import iris 


# Barometric Height Approximation 
def convertFrom_masl_to_hPa(iAltitude_m):
    iAltitude_approx_hPa=1013.25*(1-(0.0065*iAltitude_m)/288.15)**5.255 
    return iAltitude_approx_hPa                


# File/Folder-Paths (path to PT sub-domains covering geographical area of interest)
TopographyFolder="/xyz/NAME/name_v7_1/Resources/Topog"
TopographyFile_PT2=os.path.join(TopographyFolder,"TopogUMG_Mk10_PT2.pp")  
TopographyFile_PT3=os.path.join(TopographyFolder,"TopogUMG_Mk10_PT3.pp")
TopographyFile_PT6=os.path.join(TopographyFolder,"TopogUMG_Mk10_PT6.pp")
TopographyFile_PT7=os.path.join(TopographyFolder,"TopogUMG_Mk10_PT7.pp")


## Load NAME topography data 

# Get topography of sub-domain PT2. 
TopographyCubeList_PT2=iris.load(TopographyFile_PT2)
TopographyCube_PT2=TopographyCubeList_PT2[0]              
allLatitudes_PT2=TopographyCube_PT2.dim_coords[0].points
allLongitudes_PT2=TopographyCube_PT2.dim_coords[1].points

# Get topography of sub-domain PT3. 
TopographyCubeList_PT3=iris.load(TopographyFile_PT3)
TopographyCube_PT3=TopographyCubeList_PT3[0] 
allLatitudes_PT3=TopographyCube_PT3.dim_coords[0].points
allLongitudes_PT3=TopographyCube_PT3.dim_coords[1].points

# Get topography of sub-domain PT6. 
TopographyCubeList_PT6=iris.load(TopographyFile_PT6)
TopographyCube_PT6=TopographyCubeList_PT6[0] 
allLatitudes_PT6=TopographyCube_PT6.dim_coords[0].points
allLongitudes_PT6=TopographyCube_PT6.dim_coords[1].points

# Get topography of sub-domain PT7. 
TopographyCubeList_PT7=iris.load(TopographyFile_PT7)
TopographyCube_PT7=TopographyCubeList_PT7[0] 
allLatitudes_PT7=TopographyCube_PT7.dim_coords[0].points
allLongitudes_PT7=TopographyCube_PT7.dim_coords[1].points

## Define grid covering all sub-domains (PT2,3,6,7)

# Construct latitude & longitude array covering the total area  
AllLatitudes=np.append(allLatitudes_PT7,allLatitudes_PT3[2:len(allLatitudes_PT3)])
Longitudes2=allLongitudes_PT2
Longitudes2Shifted=np.empty(len(allLongitudes_PT2))
i=0
for iLon in allLongitudes_PT2:
    if iLon<=180:
        Longitudes2Shifted[i]=iLon
    if iLon>180:
        Longitudes2Shifted[i]=iLon-360
    i=i+1
Longitudes3=allLongitudes_PT3
AllLongitudes=np.append(Longitudes2Shifted,Longitudes3[2:len(Longitudes3)])

# Define grid origin and resolution. 
xGridOriginTopography=min(AllLongitudes)
yGridOriginTopography=min(AllLatitudes)
xGridResolutionTopography=AllLongitudes[1]-AllLongitudes[0] 
yGridResolutionTopography=AllLatitudes[1]-AllLatitudes[0] 

# Initialize empty array for holding surface altitude data
surfaceAltitudeData_PT2=np.empty([len(allLatitudes_PT2),len(allLongitudes_PT2)])
surfaceAltitudeData_PT3=np.empty([len(allLatitudes_PT3),len(allLongitudes_PT3)])
surfaceAltitudeData_PT6=np.empty([len(allLatitudes_PT6),len(allLongitudes_PT6)])
surfaceAltitudeData_PT7=np.empty([len(allLatitudes_PT7),len(allLongitudes_PT7)])
surfaceAltitudeData=np.empty([len(AllLatitudes),len(AllLongitudes)])

# Load topography data for each sub-domain
surfaceAltitudeData_PT2=TopographyCube_PT2.data
surfaceAltitudeData_PT3=TopographyCube_PT3.data
surfaceAltitudeData_PT6=TopographyCube_PT6.data
surfaceAltitudeData_PT7=TopographyCube_PT7.data

# Combine topography data from different sub-domain into one larger grid
surfaceAltitudeData[0:len(allLatitudes_PT6),0:len(Longitudes2Shifted)]=surfaceAltitudeData_PT6
surfaceAltitudeData[0:len(allLatitudes_PT6),len(Longitudes2Shifted):len(Longitudes2Shifted)+len(Longitudes3)-2]=surfaceAltitudeData_PT7[:,2:len(Longitudes3)]
surfaceAltitudeData[len(allLatitudes_PT6):len(allLatitudes_PT6)+len(allLatitudes_PT2)-2,0:len(Longitudes2Shifted)]=surfaceAltitudeData_PT2[2:len(allLatitudes_PT2),:]
surfaceAltitudeData[len(allLatitudes_PT6):len(allLatitudes_PT6)+len(allLatitudes_PT2)-2,len(Longitudes2Shifted):len(Longitudes2Shifted)+len(Longitudes3)-2]=surfaceAltitudeData_PT3[2:len(allLatitudes_PT2),2:len(Longitudes3)]
  
# Define 3-D field holding the terrain info (for testing alternative visualization using pressure levels) 
allVerticalLevels3DTerrainField_m=np.linspace(-100,10000,200)
allVerticalLevels3DTerrainField_hPa=np.zeros(len(allVerticalLevels3DTerrainField_m))
indexAlt=0
for iSurfaceAltitude_m in allVerticalLevels3DTerrainField_m:
    allVerticalLevels3DTerrainField_hPa[indexAlt]=convertFrom_masl_to_hPa(iSurfaceAltitude_m)
    indexAlt=indexAlt+1

# Initialize 3-D Terrain-Field 
terrainField3D=np.zeros([1,len(allVerticalLevels3DTerrainField_hPa),len(AllLatitudes),len(AllLongitudes)])
terrainField3D_m=np.zeros([1,len(allVerticalLevels3DTerrainField_hPa),len(AllLatitudes),len(AllLongitudes)])

# Get Terrain Data and Write to NetCDF file
getAndPrintTerrain=True
if getAndPrintTerrain:
    indexLat=0
    for iLat in AllLatitudes:
        print(str(iLat))
        print(str(indexLat))
        indexLon=0
        for iLon in AllLongitudes:  
            
            # Get surface altitude from topography file.
            iSurfaceAltitude_m=surfaceAltitudeData[indexLat][indexLon]

            # Convert from m above sea level to approx. pressure levels 
            iSurfaceAltitdue_hPa=convertFrom_masl_to_hPa(iSurfaceAltitude_m)

            # Find vertical levels around surface altitude and write a constant value 
            # for plotting an isosurface around the terrain in Met3D
            indexLev=0
            for iLev in allVerticalLevels3DTerrainField_hPa:
                
                # get level above 
                if indexLev<(len(allVerticalLevels3DTerrainField_hPa)-1):
                    iLev_above=allVerticalLevels3DTerrainField_hPa[indexLev+1]

                    # check if surface altitude is in this layer 
                    if (iSurfaceAltitdue_hPa<iLev) and (iSurfaceAltitdue_hPa>iLev_above):
                        
                        # write constant for plotting 
                        terrainField3D[0][indexLev][indexLat][indexLon]=1000

                        # write value of surface altitude for colouring the isosurface according to m asl
                        terrainField3D_m[0][indexLev][indexLat][indexLon]=iSurfaceAltitude_m

                        # write smaller constant in all cells below 
                        for cellsBelow in range(0,indexLev):
                            terrainField3D[0][cellsBelow][indexLat][indexLon]=10
                            terrainField3D_m[0][indexLev][indexLat][indexLon]=allVerticalLevels3DTerrainField_m[cellsBelow]
                                          
                        break

                indexLev=indexLev+1
            indexLon=indexLon+1
        indexLat=indexLat+1


    ## write netCDF file for plotting terrain in format needed for processing in Met3D
                    
    ## define filename & open for writing
    iOutFilename=os.path.join(TopographyFolder,'TerrainPT_2_3_6_7.nc')
    ncfile = netCDF4.Dataset(iOutFilename, mode="w", format="NETCDF4")

    # Write global NetCDF attributes & dimensions 
    ncfile.title="NAME (ADAQ, UKMO) terrain from topog file"
    TimeDim  = ncfile.createDimension("time", 1)
    LatDim = ncfile.createDimension("lat", len(AllLatitudes))
    LongDim = ncfile.createDimension("lon", len(AllLongitudes))
    VertDim = ncfile.createDimension("plev", len(allVerticalLevels3DTerrainField_hPa))

	## Write coordinate variables 
	# time 
    startdate=datetime.datetime(2021,3,8,00,00,00)
    timeVar = ncfile.createVariable("time", "d", ("time",))  
    timeVar.standard_name = "time"
    timeVar.long_name = "time"
    timeVar.units = startdate.strftime("hours since %Y-%m-%d %H:%M:%S")
    timeVar.axis = "T"
    timeVar.calendar = "propleptic_gregorian"

    # longitude
    lonVar = ncfile.createVariable("lon", "d", ("lon"))
    lonVar.standard_name = "longitude"
    lonVar.long_name = "longitude"
    lonVar.units = "degrees_east"
    lonVar.axis = "X"

    # latitude
    latVar = ncfile.createVariable("lat", "f", ("lat"))
    latVar.standard_name = "latitude"
    latVar.long_name = "latitude"
    latVar.units = "degrees_north"
    latVar.axis = "Y"

    # pressure levels
    presVar = ncfile.createVariable("plev", "f", ("plev"))
    presVar.standard_name = "air_pressure"
    presVar.units = "hPa"
    presVar.axis = "Z"
    presVar.positive = "up"

    # surface indication
    surfVar = ncfile.createVariable("terrainSurface", "f", ("time","plev","lat","lon"))
    surfVar.standard_name = "surfaceAltitudeIndication"
    surfVar.long_name = "surface Altitude Indication"
    surfVar.units = "--"

    # surface altitude pressure
    surfpresVar = ncfile.createVariable("terrainAltitude", "f", ("time","plev","lat","lon"))
    surfpresVar.standard_name = "surfaceAltitudePressure"
    surfpresVar.long_name = "surface Altitude Pressure"
    surfpresVar.units = "m"

    # write coordinate data  
    timeVar[:] = 0
    latVar[:]  = AllLatitudes
    lonVar[:]  = AllLongitudes
    presVar[:] = allVerticalLevels3DTerrainField_hPa

    # write terrain 
    surfVar[:,:,:,:] = terrainField3D
    surfpresVar[:,:,:,:] = terrainField3D_m

    # close
    ncfile.close()

