#------------------------------------------------------
# Method for converting NAME topography files to NetCDF files for 3D terrain visualization in ParaView 
#
# Summary:
#  --> read terrain data using Python Iris (NAME topography file; UMs pp format)
#  --> write to NetCDF file that can be processed by ParaView
#
# Disclaimer: This script provides a MINIMAL EXAMPLE that might serve as a useful first step by providing code snippets and the format data structures 
#
# @ Marcel Meyer (2022). UN Peace & Development Unit. Research conducted as part Visual Data Analysis Group, University of Hamburg in collaboration 
#                        with the Epidemiology Group at the University of Cambridge
#
#------------------


# Import libs
import datetime
import numpy as np
import pandas as pd
import os, math
import netCDF4
import socket
import datetime, dateutil, time 
import numpy as np
import iris 
from scipy.ndimage import gaussian_filter


# Function def for barometric height approximation
def convertFrom_masl_to_hPa(iAltitude_m):
    iAltitude_approx_hPa=1013.25*(1-(0.0065*iAltitude_m)/288.15)**5.255 
    return iAltitude_approx_hPa                


# File/Folder path to NAME topography files 
TopographyFolder="/xyz/NAME/name_v7_1/Resources/Topog"
TopographyFile_PT6=os.path.join(TopographyFolder,"TopogUMG_Mk10_PT6.pp")


# Get grid information for the geographical sub-domain of interest (PT6). 
TopographyCubeList_PT6=iris.load(TopographyFile_PT6)
TopographyCube_PT6=TopographyCubeList_PT6[0] 
allLatitudes_PT6=TopographyCube_PT6.dim_coords[0].points
allLongitudes_PT6=TopographyCube_PT6.dim_coords[1].points
AllLatitudes=allLatitudes_PT6
LongitudesPT6Shifted=np.empty(len(allLongitudes_PT6))
i=0
for iLon in allLongitudes_PT6:
    if iLon<=180:
        LongitudesPT6Shifted[i]=iLon
    if iLon>180:
        LongitudesPT6Shifted[i]=iLon-360
    i=i+1
AllLongitudes=LongitudesPT6Shifted

# Define grid origin and resolution. 
xGridOriginTopography=min(AllLongitudes)
yGridOriginTopography=min(AllLatitudes)
xGridResolutionTopography=AllLongitudes[1]-AllLongitudes[0] 
yGridResolutionTopography=AllLatitudes[1]-AllLatitudes[0] 

# Initialize empty array for holding topography data 
surfaceAltitudeData_PT6=np.empty([len(AllLatitudes),len(AllLongitudes)])

# Load topography data
surfaceAltitudeData_PT6=TopographyCube_PT6.data              
 

## write topography data to NetCDF file for visualization in ParaView
                
# define file & write global attributes and dimensions
iOutFilename=os.path.join("/media/bay4443@uni-hamburg.de/MM/Locusts/Test_Topography",'TerrainPT6.nc')
ncfile = netCDF4.Dataset(iOutFilename, mode="w", format="NETCDF4")
ncfile.title="NAME (ADAQ, UKMO) terrain from topog pp file"
TimeDim  = ncfile.createDimension("time", 1)
LatDim = ncfile.createDimension("lat", len(AllLatitudes))
LongDim = ncfile.createDimension("lon", len(AllLongitudes))

# Define coordinate variables (time, lat, lon , height)
# time
startdate=datetime.datetime(2021,3,8,00,00,00)
timeVar = ncfile.createVariable("time", "d", ("time",))  
timeVar.standard_name = "time"
timeVar.long_name = "time"
timeVar.units = startdate.strftime("hours since %Y-%m-%d %H:%M:%S")
timeVar.axis = "T"
timeVar.calendar = "propleptic_gregorian"

# longitude
lonVar = ncfile.createVariable("lon", "d", ("lon"))
lonVar.standard_name = "longitude"
lonVar.long_name = "longitude"
lonVar.units = "degrees_east"
lonVar.axis = "X"

# latitude
latVar = ncfile.createVariable("lat", "f", ("lat"))
latVar.standard_name = "latitude"
latVar.long_name = "latitude"
latVar.units = "degrees_north"
latVar.axis = "Y"

# surface indication
terrainVar = ncfile.createVariable("terrain", "f", ("time","lat","lon"))
terrainVar.standard_name = "terrain from NAME pp"
terrainVar.long_name = "terrain from NAME pp files"
terrainVar.units = "m asl"

# write  coordinate data

# time, lats, longs, altitude 
timeVar[:] = 0
latVar[:]  = AllLatitudes
lonVar[:]  = AllLongitudes

# write surface data
terrainVar[0,:,:] = surfaceAltitudeData_PT6

# close
ncfile.close()

