#------------------------------------------------------
# Method for converting 3-D gridded meteorological data from the UKMOs UM as used for NAME simulations from different geographical 
# sub-domains into a single data grid in a NetCDF file-format that can be processed by Met3D/ParaView for 3D visualization
#
# Summary:
# --> read UM pp files using the python iris package
# --> combine met data from different spatial sub-domains into a common grid 
# --> write met data to netCDF file
#
# Disclaimer: This script provides a MINIMAL EXAMPLE that might serve as a useful first step by providing code snippets and the format data structures 
#
# @ Marcel Meyer (2022). UN Peace & Development Unit. Research conducted as part Visual Data Analysis Group, University of Hamburg in collaboration 
#                        with the Epidemiology Group at the University of Cambridge
#
#------------------


import copy, time, os, glob, sys, re, math, csv
import datetime, dateutil, time 
import numpy as np
import netCDF4
import socket
import iris 


# file/folder-paths to UM meteo data used as input for NAME simulations
ProjectFolderPath="/xyz/"
MetDataFolder=os.path.join(ProjectFolderPath,"UM_TestData")

# def. time-domain for sample test data 
startdate=datetime.datetime(2021,3,2,00,00,00)
enddate=datetime.datetime(2021,3,2,21,00,00)
DeltaT_Hours=3
Delta_t=datetime.timedelta(hours=3)

# load & convert meteo. data in time-domain of interest
iDate=startdate
while iDate<enddate:
    
    # set date and keep track
    datestring=iDate.strftime('%Y%m%d%H%M')
    print("processing: ", datestring)          
            
    # def. list of met-data filenames 
	ListFilenamesI=[os.path.join(MetDataFolder,'MO'+datestring+'.UMG_Mk10_I_L59PT2.pp'),
        os.path.join(MetDataFolder,'MO'+datestring+'.UMG_Mk10_I_L59PT3.pp'),
        os.path.join(MetDataFolder,'MO'+datestring+'.UMG_Mk10_I_L59PT6.pp'),
        os.path.join(MetDataFolder,'MO'+datestring+'.UMG_Mk10_I_L59PT7.pp')]
    print('processing: ',ListFilenamesI)
  
    # load meta-data to Iris cube 
    metVarNamesI=['air_pressure','x_wind','y_wind',
    'upward_air_velocity','surface_air_pressure',
    'air_temperature']
    iCubeList=iris.load(ListFilenamesI,metVarNamesI)
 
    # separate the met-variables of interest into different cube lists 
    # note: the selection of variables via hardcoded indexing works for this minimal example, but selection by name should be considered
    # note: over longer time-period the met. data is not always consistent, which leads to, e.g., different cube-lists, so if longer time-periods are of interest, then careful handling is required
    iCubeListPressure=iris.cube.CubeList([iCubeList[0],iCubeList[1],iCubeList[2],iCubeList[3]])
    iCubeListXWinds=iris.cube.CubeList([iCubeList[4],iCubeList[5],iCubeList[6],iCubeList[7]])
    iCubeListYWinds=iris.cube.CubeList([iCubeList[8],iCubeList[9],iCubeList[10],iCubeList[11]])
    iCubeListZWinds=iris.cube.CubeList([iCubeList[12],iCubeList[13],iCubeList[14],iCubeList[15]])
    iCubeListSurfacePressure=iris.cube.CubeList([iCubeList[16],iCubeList[17],iCubeList[18],iCubeList[19]])
    iCubeListAirTemperature=iris.cube.CubeList([iCubeList[24],iCubeList[25],iCubeList[26],iCubeList[27]])
    
    # combine met data from different spatial sub-domains
	# note: unfortunately, the built-in iris method for concatenating data fails here because there are small overlaps
    # in the latitudes covered by the different PT sub-domains (duplicated lats) and because the longitudes are defined with a different origin in different PT sub-domains
    # work-around: manual concatenating of sub-domains, 
	
	# construct latitude array for grid covering all PT sub-domains 
	# note: PT 6 and 7 cover latitudes -25 to 25 and PT 2 and 3 cover laitudes > 25
    # note: there is an overlap of two latitudes at the boundary between PT2 and PT6 as well as PT3 and PT7
    # note: avoid duplicates
    LatitudesPT6=iCubeListPressure[2].dim_coords[1]    
    LatitudesPT2=iCubeListPressure[0].dim_coords[1]    
    AllLatitudes=np.append(LatitudesPT6.points,LatitudesPT2.points[2:len(LatitudesPT2.points)])
    
    # construct longitude array for grid covering all PT sub-domains
    # note: PT 2 and 6 cover longitudes -45 to 45 and PT 3 and 7 cover longitudes 45 - 135
    # note: PT 2 and 6 are defined using longitudes running from 315 until 414 
    # note: PT 3 and 7 are defined using longitudes running as expected from 45 to 135
    # note: there is an overlap of two longitudes at the boundary between PT2 and PT3 as well as PT6 and PT7
    # note: for merging, shift the longitudes to a common frame, then append counting duplicates once
    Longitudes2=iCubeListPressure[0].dim_coords[2]
    Longitudes2Shifted=np.empty(len(Longitudes2.points))
    i=0
    for iLon in Longitudes2.points:
        if iLon<=180:
            Longitudes2Shifted[i]=iLon
        if iLon>180:
            Longitudes2Shifted[i]=iLon-360
        i=i+1
    Longitudes3=iCubeListPressure[1].dim_coords[2]
    AllLongitudes=np.append(Longitudes2Shifted,Longitudes3.points[2:len(Longitudes3.points)])

	# construct altitude array
    AllAltitudes=iCubeListPressure[0].dim_coords[0] 

    # initialize empty arrays for storing the met-data
    PressureData=np.empty([1,len(AllAltitudes.points),len(AllLatitudes),len(AllLongitudes)])
    XWindData=np.empty([1,len(AllAltitudes.points),len(AllLatitudes),len(AllLongitudes)])
    YWindData=np.empty([1,len(AllAltitudes.points),len(AllLatitudes),len(AllLongitudes)])
    ZWindData=np.empty([1,len(AllAltitudes.points),len(AllLatitudes),len(AllLongitudes)])
    SurfacePressureData=np.empty([1,len(AllLatitudes),len(AllLongitudes)])
    AirTData=np.empty([1,len(AllAltitudes.points),len(AllLatitudes),len(AllLongitudes)])
    
    # fill pressure array with data from different PT sub-domains 
    # bottom left (PT 6) 
    PressureData[0,:,0:len(LatitudesPT6.points),0:len(Longitudes2Shifted)]=iCubeListPressure[2].data
    # bottom right (PT 7) 
    PressureData[0,:,0:len(LatitudesPT6.points),len(Longitudes2Shifted):len(Longitudes2Shifted)+len(Longitudes3.points)-2]=iCubeListPressure[3].data[:,:,2:len(Longitudes3.points)]
    # top left (PT2)
    PressureData[0,:,len(LatitudesPT6.points):len(LatitudesPT6.points)+len(LatitudesPT2.points)-2,0:len(Longitudes2Shifted)]=iCubeListPressure[0].data[:,2:len(LatitudesPT2.points),:]
    # top right (PT 3)
    PressureData[0,:,len(LatitudesPT6.points):len(LatitudesPT6.points)+len(LatitudesPT2.points)-2,len(Longitudes2Shifted):len(Longitudes2Shifted)+len(Longitudes3.points)-2]=iCubeListPressure[1].data[:,2:len(LatitudesPT2.points),2:len(Longitudes3.points)]
    
    # fill x-wind array with data from different PT sub-domains 
    XWindData[0,:,0:len(LatitudesPT6.points),0:len(Longitudes2Shifted)]=iCubeListXWinds[2].data
    XWindData[0,:,0:len(LatitudesPT6.points),len(Longitudes2Shifted):len(Longitudes2Shifted)+len(Longitudes3.points)-2]=iCubeListXWinds[3].data[:,:,2:len(Longitudes3.points)]
    XWindData[0,:,len(LatitudesPT6.points):len(LatitudesPT6.points)+len(LatitudesPT2.points)-2,0:len(Longitudes2Shifted)]=iCubeListXWinds[0].data[:,2:len(LatitudesPT2.points),:]
    XWindData[0,:,len(LatitudesPT6.points):len(LatitudesPT6.points)+len(LatitudesPT2.points)-2,len(Longitudes2Shifted):len(Longitudes2Shifted)+len(Longitudes3.points)-2]=iCubeListXWinds[1].data[:,2:len(LatitudesPT2.points),2:len(Longitudes3.points)]
    
    # fill y-wind array with data from different PT sub-domains
    # note: y-wind latitudes are shifted by 1/2 grid-cell as it is defined in cell center; there is 1 more entry per cube and 3 overlaps
    # note: ignore this for the minimal example of visualizations produced here, because sub-grid cell differences are not visible anyways
    YWindData[0,:,0:len(LatitudesPT6.points),0:len(Longitudes2Shifted)]=iCubeListYWinds[2].data[:,0:len(LatitudesPT6.points),:]
    YWindData[0,:,0:len(LatitudesPT6.points),len(Longitudes2Shifted):len(Longitudes2Shifted)+len(Longitudes3.points)-2]=iCubeListYWinds[3].data[:,0:len(LatitudesPT6.points),2:len(Longitudes3.points)]
    YWindData[0,:,len(LatitudesPT6.points):len(LatitudesPT6.points)+len(LatitudesPT2.points)-2,0:len(Longitudes2Shifted)]=iCubeListYWinds[0].data[:,3:len(LatitudesPT2.points)+1,:]
    YWindData[0,:,len(LatitudesPT6.points):len(LatitudesPT6.points)+len(LatitudesPT2.points)-2,len(Longitudes2Shifted):len(Longitudes2Shifted)+len(Longitudes3.points)-2]=iCubeListYWinds[1].data[:,3:len(LatitudesPT2.points)+1,2:len(Longitudes3.points)]
    
    # fill z-wind array with data from different PT sub-domains
    ZWindData[0,:,0:len(LatitudesPT6.points),0:len(Longitudes2Shifted)]=iCubeListZWinds[2].data
    ZWindData[0,:,0:len(LatitudesPT6.points),len(Longitudes2Shifted):len(Longitudes2Shifted)+len(Longitudes3.points)-2]=iCubeListZWinds[3].data[:,:,2:len(Longitudes3.points)]
    ZWindData[0,:,len(LatitudesPT6.points):len(LatitudesPT6.points)+len(LatitudesPT2.points)-2,0:len(Longitudes2Shifted)]=iCubeListZWinds[0].data[:,2:len(LatitudesPT2.points),:]
    ZWindData[0,:,len(LatitudesPT6.points):len(LatitudesPT6.points)+len(LatitudesPT2.points)-2,len(Longitudes2Shifted):len(Longitudes2Shifted)+len(Longitudes3.points)-2]=iCubeListZWinds[1].data[:,2:len(LatitudesPT2.points),2:len(Longitudes3.points)]
    
    # fill surface pressure array with data from different PT sub-domains
    SurfacePressureData[0,0:len(LatitudesPT6.points),0:len(Longitudes2Shifted)]=iCubeListSurfacePressure[2].data
    SurfacePressureData[0,0:len(LatitudesPT6.points),len(Longitudes2Shifted):len(Longitudes2Shifted)+len(Longitudes3.points)-2]=iCubeListSurfacePressure[3].data[:,2:len(Longitudes3.points)]
    SurfacePressureData[0,len(LatitudesPT6.points):len(LatitudesPT6.points)+len(LatitudesPT2.points)-2,0:len(Longitudes2Shifted)]=iCubeListSurfacePressure[0].data[2:len(LatitudesPT2.points),:]
    SurfacePressureData[0,len(LatitudesPT6.points):len(LatitudesPT6.points)+len(LatitudesPT2.points)-2,len(Longitudes2Shifted):len(Longitudes2Shifted)+len(Longitudes3.points)-2]=iCubeListSurfacePressure[1].data[2:len(LatitudesPT2.points),2:len(Longitudes3.points)]
    
    # fill temperature array with data from different PT sub-domains
    AirTData[0,:,0:len(LatitudesPT6.points),0:len(Longitudes2Shifted)]=iCubeListAirTemperature[2].data
    AirTData[0,:,0:len(LatitudesPT6.points),len(Longitudes2Shifted):len(Longitudes2Shifted)+len(Longitudes3.points)-2]=iCubeListAirTemperature[3].data[:,:,2:len(Longitudes3.points)]
    AirTData[0,:,len(LatitudesPT6.points):len(LatitudesPT6.points)+len(LatitudesPT2.points)-2,0:len(Longitudes2Shifted)]=iCubeListAirTemperature[0].data[:,2:len(LatitudesPT2.points),:]
    AirTData[0,:,len(LatitudesPT6.points):len(LatitudesPT6.points)+len(LatitudesPT2.points)-2,len(Longitudes2Shifted):len(Longitudes2Shifted)+len(Longitudes3.points)-2]=iCubeListAirTemperature[1].data[:,2:len(LatitudesPT2.points),2:len(Longitudes3.points)]
    
    
    # write data into netCDF file in format required for processing in Met3D
                                
    # def. filename & write global attributes and dimensions
    iOutFilename=os.path.join(MetDataFolder,'NCFiles/MO'+datestring+'.UMG_Mk10_IM_L59PT2367.nc')
    ncfile = netCDF4.Dataset(iOutFilename, mode="w", format="NETCDF4")
    ncfile.title=list(iCubeListPressure[0].attributes.values())[0]
    TimeDim  = ncfile.createDimension("time", 1)
    LatDim = ncfile.createDimension("lat", len(AllLatitudes))
    LongDim = ncfile.createDimension("lon", len(AllLongitudes))
    VertDim = ncfile.createDimension("height", len(AllAltitudes.points))
    
    # write coordinate variables (time, lat, lon , height)
    
	# time 
	timeVar = ncfile.createVariable("time", "d", ("time",))  
    timeVar.standard_name = "time"
    timeVar.long_name = "time"
    timeVar.units = iDate.strftime("hours since %Y-%m-%d %H:%M:%S")
    timeVar.axis = "T"
    timeVar.calendar = "propleptic_gregorian"

    # longitude
    lonVar = ncfile.createVariable("lon", "d", ("lon"))
    lonVar.standard_name = "longitude"
    lonVar.long_name = "longitude"
    lonVar.units = "degrees_east"
    lonVar.axis = "X"

    # latitude
    latVar = ncfile.createVariable("lat", "f", ("lat"))
    latVar.standard_name = "latitude"
    latVar.long_name = "latitude"
    latVar.units = "degrees_north"
    latVar.axis = "Y"

    # height
    HeightVar = ncfile.createVariable("height", "f", ("height"))
    HeightVar.standard_name = "height"
    HeightVar.long_name = "generalized_height"
    HeightVar.axis = "Z"

    # def met data variables
	
	# aux pressure
    presVar = ncfile.createVariable("pres", "f", ("time","height","lat","lon"))
    presVar.standard_name = "Pressure"
    presVar.units = "Pa"

    # surface pressure
    surpresVar = ncfile.createVariable("surfacepres", "f", ("time","lat","lon"))
    surpresVar.standard_name = "surface_air_pressure"
    surpresVar.units = "Pa"

    # x-winds
    XWindVar = ncfile.createVariable("u", "f", ("time","height","lat","lon"))
    XWindVar.standard_name = "eastwards_wind"
    XWindVar.long_name = "x_wind (zonal)"
    XWindVar.units = iCubeListXWinds[0].units.definition

	# y-winds 
    YWindVar = ncfile.createVariable("v", "f", ("time","height","lat","lon"))
    YWindVar.long_name = "y_wind (meridional)"
    YWindVar.standard_name = "northward_wind"
    YWindVar.units = iCubeListYWinds[0].units.definition

	# z-winds 
    ZWindVar = ncfile.createVariable("w", "f", ("time","height","lat","lon"))
    ZWindVar.standard_name = iCubeListZWinds[0].standard_name
    ZWindVar.long_name = "vertical_wind (upwards)"
    ZWindVar.units = iCubeListZWinds[0].units.definition

    # temperature
    TVar = ncfile.createVariable("Temp", "f", ("time","height","lat","lon"))
    TVar.standard_name = iCubeListAirTemperature[0].standard_name
    TVar.long_name = "temperature"
    TVar.units = "K"
    

    # write to nc arrays 
    timeVar[:] = 0   # time-difference
    latVar[:]  = AllLatitudes
    lonVar[:]  = AllLongitudes
    HeightVar[:] = AllAltitudes.points
    presVar[0,:,:,:] = PressureData
    surpresVar[0,:,:] = SurfacePressureData
    XWindVar[0,:,:,:] = XWindData
    YWindVar[0,:,:,:] = YWindData
    ZWindVar[0,:,:,:] = ZWindData      
    TVar[0,:,:,:] = AirTData
    
    # close file 
    ncfile.close()

    # iterate time
    iDate=iDate+Delta_t


