#------------------------------------------------------
# Method for converting 3-D gridded meteorological data from the UKMOs UM as used for NAME simulations from a single geographical 
# sub-domains (PT domain) into a NetCDF file-format that can be processed by Met3D/ParaView for 3D visualization
#
# Summary:
# --> read UM pp files using the python iris package
# --> write met data to netCDF file
#
# Disclaimer: This script provides a MINIMAL EXAMPLE that might serve as a useful first step by providing code snippets and the format data structures 
#
# @ Marcel Meyer (2022). UN Peace & Development Unit. Research conducted as part Visual Data Analysis Group, University of Hamburg in collaboration 
#                        with the Epidemiology Group at the University of Cambridge
#
#------------------


import copy, time, os, glob, sys, re, math, csv
import datetime, dateutil, time 
import numpy as np
import netCDF4
import socket
import iris 
import time 


# file/folder-paths to met-data
ProjectFolderPath="/xyz/"
MetDataFolder=os.path.join(ProjectFolderPath,"TestData")

# def. time-domain for sample test data 
startdate=datetime.datetime(2020,2,7,00,00,00)
enddate=datetime.datetime(2020,2,8,21,00,00)
DeltaT_Hours=3
Delta_t=datetime.timedelta(hours=3)
print("Start-Date:",startdate)
print("Delta T:",Delta_t)

# load data from UKMO pp files and write to netCDF file
iDate=startdate
while iDate<enddate:
    
    # set date and keep track
    datestring=iDate.strftime('%Y%m%d%H%M')
    print("processing: ", datestring)          
            
    # def. list of filenames 
    ListFilenamesI=[os.path.join(MetDataFolder,'MO'+datestring+'.UMG_Mk10_I_L59PT6.pp')]
    print('processing: ',ListFilenamesI)
  
    # load all meta-data into iris cube list
    metVarNamesI=['air_pressure','x_wind','y_wind',
    'upward_air_velocity','surface_air_pressure',
    'air_temperature']
    iCubeList=iris.load(ListFilenamesI,metVarNamesI)
 
    # construct one cube list per met variable 
    iCubeListPressure=iris.cube.CubeList([iCubeList[0]])
    iCubeListXWinds=iris.cube.CubeList([iCubeList[1]])
    iCubeListYWinds=iris.cube.CubeList([iCubeList[2]])
    iCubeListZWinds=iris.cube.CubeList([iCubeList[3]])
    iCubeListSurfacePressure=iris.cube.CubeList([iCubeList[4]])
    iCubeListAirTemperature=iris.cube.CubeList([iCubeList[6]])
    
    # construct target grid (lat,lon,height)
	
	# latitudes
    LatitudesPT6=iCubeListPressure[0].dim_coords[1]    
    LongitudesPT6=iCubeListPressure[0].dim_coords[2]
    
    # longitudes
    LongitudesPT6Shifted=np.empty(len(LongitudesPT6.points))
    i=0
    for iLon in LongitudesPT6.points:
        if iLon<=180:
            LongitudesPT6Shifted[i]=iLon
        if iLon>180:
            LongitudesPT6Shifted[i]=iLon-360
        i=i+1
   
    # height
    AllAltitudes=iCubeListPressure[0].dim_coords[0] # identical for all PT domains

    # initialize empty data arrays
    PressureData=np.empty([1,len(AllAltitudes.points),len(LatitudesPT6.points),len(LongitudesPT6Shifted)])
    XWindData=np.empty([1,len(AllAltitudes.points),len(LatitudesPT6.points),len(LongitudesPT6Shifted)])
    YWindData=np.empty([1,len(AllAltitudes.points),len(LatitudesPT6.points),len(LongitudesPT6Shifted)])
    ZWindData=np.empty([1,len(AllAltitudes.points),len(LatitudesPT6.points),len(LongitudesPT6Shifted)])
    SurfacePressureData=np.empty([1,len(LatitudesPT6.points),len(LongitudesPT6Shifted)])
    AirTData=np.empty([1,len(AllAltitudes.points),len(LatitudesPT6.points),len(LongitudesPT6Shifted)])
    
    # fill with met data 
    PressureData[0,:,:,:]=iCubeListPressure[0].data[:,0:len(LatitudesPT6.points),0:len(LongitudesPT6Shifted)]
    XWindData[0,:,:,:]=iCubeListXWinds[0].data[:,0:len(LatitudesPT6.points),0:len(LongitudesPT6Shifted)]
    YWindData[0,:,:,:]=iCubeListYWinds[0].data[:,0:len(LatitudesPT6.points),0:len(LongitudesPT6Shifted)]
    ZWindData[0,:,:,:]=iCubeListZWinds[0].data[:,0:len(LatitudesPT6.points),0:len(LongitudesPT6Shifted)]
    SurfacePressureData[0,:,:]=iCubeListSurfacePressure[0].data[0:len(LatitudesPT6.points),0:len(LongitudesPT6Shifted)]
    AirTData[0,:,:,:]=iCubeListAirTemperature[0].data[:,0:len(LatitudesPT6.points),0:len(LongitudesPT6Shifted)]
        
    # write data into netCDF file in format required by Met3D
                                
    # def. filename, open and write attributes & dimension 
    iOutFilename=os.path.join(MetDataFolder,'NCFiles/MO'+datestring+'.UMG_Mk10_IM_L59PT6.nc')
    ncfile = netCDF4.Dataset(iOutFilename, mode="w", format="NETCDF4")
    ncfile.title=list(iCubeListPressure[0].attributes.values())[0]
    TimeDim  = ncfile.createDimension("time", 1)
    LatDim = ncfile.createDimension("lat", len(LatitudesPT6.points))
    LongDim = ncfile.createDimension("lon", len(LongitudesPT6Shifted))
    VertDim = ncfile.createDimension("height", len(AllAltitudes.points))
    
    # def. coordinate variables (time, lat, lon , height).
    
	# time
    timeVar = ncfile.createVariable("time", "d", ("time",))  
    timeVar.standard_name = "time"
    timeVar.long_name = "time"
    timeVar.units = iDate.strftime("hours since %Y-%m-%d %H:%M:%S")
    timeVar.axis = "T"
    timeVar.calendar = "propleptic_gregorian"

    # longitude
    lonVar = ncfile.createVariable("lon", "d", ("lon"))
    lonVar.standard_name = "longitude"
    lonVar.long_name = "longitude"
    lonVar.units = "degrees_east"
    lonVar.axis = "X"

    # latitude
    latVar = ncfile.createVariable("lat", "f", ("lat"))
    latVar.standard_name = "latitude"
    latVar.long_name = "latitude"
    latVar.units = "degrees_north"
    latVar.axis = "Y"

    # height
    HeightVar = ncfile.createVariable("height", "f", ("height"))
    HeightVar.standard_name = "height"
    HeightVar.long_name = "generalized_height"
    HeightVar.axis = "Z"

    # def. met variables 
	
	# aux pressure
    presVar = ncfile.createVariable("pres", "f", ("time","height","lat","lon"))
    presVar.standard_name = "Pressure"
    presVar.units = "Pa"

    # surface pressure
    surpresVar = ncfile.createVariable("surfacepres", "f", ("time","lat","lon"))
    surpresVar.standard_name = "surface_air_pressure"
    surpresVar.units = "Pa"

    # wind data
    XWindVar = ncfile.createVariable("u", "f", ("time","height","lat","lon"))
    XWindVar.standard_name = "eastwards_wind"
    XWindVar.long_name = "x_wind (zonal)"
    XWindVar.units = iCubeListXWinds[0].units.definition

    YWindVar = ncfile.createVariable("v", "f", ("time","height","lat","lon"))
    YWindVar.long_name = "y_wind (meridional)"
    YWindVar.standard_name = "northward_wind"
    YWindVar.units = iCubeListYWinds[0].units.definition

    ZWindVar = ncfile.createVariable("w", "f", ("time","height","lat","lon"))
    ZWindVar.standard_name = iCubeListZWinds[0].standard_name
    ZWindVar.long_name = "vertical_wind (upwards)"
    ZWindVar.units = iCubeListZWinds[0].units.definition

    # temperature
    TVar = ncfile.createVariable("Temp", "f", ("time","height","lat","lon"))
    TVar.standard_name = iCubeListAirTemperature[0].standard_name
    TVar.long_name = "temperature"
    TVar.units = "K"   

    # write data to nc arrays
    timeVar[:] = 0   
    latVar[:]  = LatitudesPT6.points
    lonVar[:]  = LongitudesPT6Shifted
    HeightVar[:] = AllAltitudes.points
    presVar[0,:,:,:] = PressureData
    surpresVar[0,:,:] = SurfacePressureData
    XWindVar[0,:,:,:] = XWindData
    YWindVar[0,:,:,:] = YWindData
    ZWindVar[0,:,:,:] = ZWindData      
    TVar[0,:,:,:] = AirTData   
 
    # close file 
    ncfile.close()

    # iterate time 
    iDate=iDate+Delta_t

